# Serene Programming Language
#
# Copyright (c) 2019-2024 Sameer Rahmani <lxsameer@gnu.org>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
{
  description = "Serene programming language";

  #inputs.nixpkgs.url = "github:NixOS/nixpkgs/bcb1a9c7e1d8568c5e58316fe3254eb8f4455439";
  #inputs.nixpkgs.url = "github:lxsameer/nixpkgs/e1f7865bce4d52d30dd1d61e79798ee2765cc2b0";
  inputs.nixpkgs.url = "/home/lxsameer/src/nixpkgs/";
  inputs.git-hooks.url = "github:cachix/git-hooks.nix";
  inputs.flake-utils.url = "github:numtide/flake-utils";

  outputs = { self, nixpkgs, flake-utils, git-hooks, ... }:
    flake-utils.lib.eachDefaultSystem (system:
      let
        version = "1.0.0";
        # Most of these overlays are do to bugs and problems
        # in upstream nixpkgs. But thanks to their design
        # We can fix them using these overlays and contribuete
        # them upstream little by little.
        overlays = [
          (final: prev: {
            p11-kit = prev.p11-kit.overrideAttrs
              (old: { patches = [ ./nix/patches/p11-kit_skip_test.patch ]; });

            cpio = prev.cpio.overrideAttrs (old: {
              nativeBuildInputs = [ prev.autoreconfHook ];
              NIX_CFLAGS_COMPILE = "-Wno-implicit-function-declaration";
            });

            libedit = prev.libedit.overrideAttrs (old: {
              # Musl is ISO 10646 compliant but doesn't define __STDC_ISO_10646__ we need to do it ourselves
              NIX_CFLAGS_COMPILE = "-D__STDC_ISO_10646__=201103L";
            });

            elfutils = prev.elfutils.overrideAttrs (old: {
              # libcxx does not have __cxa_demangle
              configureFlags = old.configureFlags ++ [ "--disable-demangler" ];
            });

            ccache = prev.ccache.overrideAttrs (old: {
              nativeBuildInputs = old.nativeBuildInputs ++ [ final.elfutils ];
            });

            # We don't need systemd at all
            util-linux = prev.util-linux.override { systemdSupport = false; };

            # libpam exmaples use glibc. We need to disable them
            linux-pam = prev.linux-pam.overrideAttrs (old: {
              postConfigure = ''
                sed 's/examples//' -i Makefile
              '';
            });

            #=============================================================
            # Since we're using lld-18, and --no-undefined-version is the
            # default in lld-18. We need to explicitly turn it off for
            # these problematic packages untill they fix it upstream.
            libgcrypt = prev.libgcrypt.overrideAttrs (old: {
              NIX_LDFLAGS = if prev.stdenv.cc.isClang then
                [ "--undefined-version" ]
              else
                [ ];
            });
            libxcrypt = prev.libxcrypt.overrideAttrs (old: {
              NIX_LDFLAGS = if prev.stdenv.cc.isClang then
                [ "--undefined-version" ]
              else
                [ ];
            });
            ncurses = prev.ncurses.overrideAttrs (old: {
              NIX_LDFLAGS = if prev.stdenv.cc.isClang then
                [ "--undefined-version" ]
              else
                [ ];
            });

            libbsd = prev.libbsd.overrideAttrs (old: {
              NIX_LDFLAGS = if prev.stdenv.cc.isClang then
                [ "--undefined-version" ]
              else
                [ ];
            });

            binutils = prev.binutils.overrideAttrs (old: {
              NIX_LDFLAGS = [
                "-Wl,--undefined-version"
              ] # if prev.stdenv.cc.isClang
              # then
              # else []
              ;
            });

            libidn2 = prev.libidn2.overrideAttrs (old: {
              # old.NIX_LDFLAGS ++
              NIX_LDFLAGS = if prev.stdenv.cc.isClang then
                [ "--undefined-version" ]
              else
                [ ];
            });

            nghttp2 = prev.nghttp2.overrideAttrs
              (old: { buildInputs = old.buildInputs ++ [ final.zlib ]; });

            #==============================================================

            iwyu = (prev.include-what-you-use.overrideAttrs (old:
              let version = "0.22";
              in {
                inherit version;

                src = prev.fetchurl {
                  url =
                    "${old.meta.homepage}/downloads/${old.pname}-${version}.src.tar.gz";
                  hash = "sha256-hZB0tGHqS4MlpzQYwgfKM7XmVmsI5rWH65FkQWVppt0=";
                };
                cmakeFlags =
                  [ "-DCMAKE_PREFIX_PATH=${prev.llvmPackages_18.llvm.dev}" ];

              })).override {
                llvmPackages = prev.__splicedPackages.llvmPackages_18;
              };
          })
        ];

        utils = import ./nix/utils.nix { inherit nixpkgs; };

        # Create a package set based on the build system
        pkgs_set = utils.get_pkgs system overlays;
        hostPkgs = pkgs_set.host;
        pkgs = pkgs_set.target;

        # Create a stdenv based on LLVM
        stdenv = pkgs.stdenvAdapters.overrideCC pkgs.stdenv
          pkgs.llvmPackages_18.clangUseLLVM;

        gc = pkgs.callPackage ./nix/boehmgc.nix { inherit stdenv; };

        zlib' = pkgs.zlib-ng.overrideAttrs (old: {
          cmakeFlags = [
            "-DCMAKE_INSTALL_PREFIX=/"
            "-DBUILD_SHARED_LIBS=OFF"
            "-DINSTALL_UTILS=ON"
            "-DZLIB_COMPAT=ON"
          ];
        });

        # By default llvm adds zlib to `propagatedBuildInputs` which means any
        # package that uses llvm will indirectly depends on zlib. And since
        # by default that zlib is built as a shared lib (since our packageset
        # is not static), We can't statically link to it. So, we just replace
        # that zlib with our override of zlib-ng
        clang' = stdenv.cc.overrideAttrs (old: {
          propagatedBuildInputs = [ stdenv.cc.bintools ] ++ [ zlib' ];
        });
        llvm = pkgs.llvmPackages_18.llvm.overrideAttrs
          (old: { propagatedBuildInputs = [ zlib' ]; });

        # This is the actual stdenv that we need to use anywhere else
        stdenv' = pkgs.stdenvAdapters.overrideCC pkgs.stdenv clang';

        nativeBuildToolsDeps = (with hostPkgs; [ cmake ninja ccache ]);

        buildToolsDeps =
          (with pkgs; [ llvm llvmPackages_18.mlir llvmPackages_18.clang iwyu ]);

        buildDeps = (with pkgs; [
          gc
          zlib'
          llvm
          llvmPackages_18.mlir
          llvmPackages_18.clang
        ]);

        testDeps = (with hostPkgs; [ gtest gmock gbenchmark ]);

      in {
        inherit pkgs;
        devShells.default = (pkgs.mkShell.override { stdenv = stdenv'; }) {
          inherit (self.checks.${system}.git-hook-check) shellHook;

          nativeBuildInputs = nativeBuildToolsDeps ++ buildToolsDeps;
          buildInputs = buildDeps ++ testDeps
            ++ self.checks.${system}.git-hook-check.enabledPackages;
          CPP_LS = "serene-clangd";
        };

        packages.devshell = stdenv'.mkDerivation {
          inherit version;
          name = "devshell";

          doUnpack = false;
          doCheck = false;
          nativeBuildInputs = nativeBuildToolsDeps ++ buildToolsDeps;
          buildInputs = buildDeps ++ testDeps;
        };

        checks = {
          git-hook-check = git-hooks.lib.${system}.run {
            src = ./.;
            hooks = {
              nixfmt.enable = true;
              clang-format = {
                enable = true;
                types_or = hostPkgs.lib.mkForce [ "c" "c++" ];
              };
              shellcheck.enable = true;
              cmake-format.enable = true;

            };
          };
        };

      });
}
